from lxml import etree as ET
import os
import logging
from . import utils, core
from pathlib import Path

# Get access to logger
log = logging.getLogger("ptxlogger")

# generate latex-image assets


def latex_image(
    ptxfile: Path,
    pub_file: Path,
    output: Path,
    params,
    target_format,
    xmlid_root,
    pdf_method,
    all_formats=False,
):
    # Dictionary of formats for images based on target
    formats = {
        "pdf": None,
        "latex": None,
        "html": ["svg"],
        "epub": ["svg"],
        "kindle": ["png"],
    }
    # set overwrite formats to all when appropriate
    if all_formats:
        formats[target_format] = {key: ["all"] for key in formats[target_format]}
    # We assume passed paths are absolute.
    # set images directory
    # parse source so we can check for latex-image.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if (
        len(source_xml.xpath("/pretext/*[not(docinfo)]//latex-image")) > 0
        and formats[target_format] is not None
    ):
        image_output = (output / "latex-image").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating latex-images\n\n")
        # Check for external requirements
        utils.check_asset_execs("latex-image", formats[target_format])
        # call pretext-core's latex image module:
        with utils.working_directory(Path()):
            for outformat in formats[target_format]:
                try:
                    core.latex_image_conversion(
                        xml_source=ptxfile,
                        pub_file=pub_file.as_posix(),
                        stringparams=params,
                        xmlid_root=xmlid_root,
                        dest_dir=image_output.as_posix(),
                        outformat=outformat,
                        method=pdf_method,
                    )
                except Exception as e:
                    log.error(e)
                    log.debug("Exception info:\n##################\n", exc_info=True)
                    log.info("##################")
                    log.error(
                        "Failed to generate some latex-image elements.  Check your source and partial output to diagnose the issue."
                    )
                    log.warning("Continuing...")
    else:
        log.info("Note: No latex-image elements found.")


# generate sageplot assets


def sageplot(
    ptxfile: Path,
    pub_file: Path,
    output: Path,
    params,
    target_format,
    xmlid_root,
    all_formats=False,
):
    # Dictionary of formats for images based on target
    formats = {
        "pdf": ["pdf", "png"],
        "latex": ["pdf", "png"],
        "html": ["html", "svg"],
        "epub": ["svg"],
        "kindle": ["png"],
    }
    # set overwrite formats to all when appropriate
    if all_formats:
        formats[target_format] = {key: ["all"] for key in formats[target_format]}
    # We assume passed paths are absolute.
    # set images directory
    # parse source so we can check for sageplot.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if (
        len(source_xml.xpath("/pretext/*[not(docinfo)]//sageplot")) > 0
        and formats[target_format] is not None
    ):
        image_output = (output / "sageplot").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating sageplot images\n\n")
        # Check for external requirements
        utils.check_asset_execs("sageplot", formats[target_format])
        with utils.working_directory(Path()):
            try:
                for outformat in formats[target_format]:
                    core.sage_conversion(
                        xml_source=ptxfile,
                        pub_file=pub_file.as_posix(),
                        stringparams=params,
                        xmlid_root=xmlid_root,
                        dest_dir=image_output.as_posix(),
                        outformat=outformat,
                    )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error(
                    "Failed to generate some sageplot elements.  Check your source and partial output to diagnose the issue."
                )
                log.warning("Continuing...")
    else:
        log.info("Note: No sageplot elements found.")


# generate asymptote assets


def asymptote(
    ptxfile: Path,
    pub_file: Path,
    output: Path,
    params,
    target_format,
    xmlid_root,
    all_formats=False,
):
    # Dictionary of formats for images based on target
    formats = {
        "pdf": ["pdf"],
        "latex": ["pdf"],
        "html": ["html"],
        "epub": ["svg"],
        "kindle": ["png"],
    }
    # set overwrite formats to all when appropriate
    if all_formats:
        formats[target_format] = {key: ["all"] for key in formats[target_format]}
    # We assume passed paths are absolute.
    # parse source so we can check for asymptote.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if (
        len(source_xml.xpath("/pretext/*[not(docinfo)]//asymptote")) > 0
        and formats[target_format] is not None
    ):
        image_output = (output / "asymptote").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating asymptote images\n\n")
        with utils.working_directory(Path()):
            try:
                for outformat in formats[target_format]:
                    core.asymptote_conversion(
                        xml_source=ptxfile,
                        pub_file=pub_file.as_posix(),
                        stringparams=params,
                        xmlid_root=xmlid_root,
                        dest_dir=image_output.as_posix(),
                        outformat=outformat,
                        method="server",
                    )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error(
                    "Failed to generate some asymptote elements. Check your source and partial output to diagnose the issue."
                )
                log.warning("Continuing...")
    else:
        log.info("Note: No asymptote elements found.")


# generate interactive preview assets


def interactive(ptxfile: Path, pub_file: Path, output: Path, params, xmlid_root):
    # First verify that playwright has dependencies installed:
    utils.playwright_install()
    # We assume passed paths are absolute.
    # parse source so we can check for interactives.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if len(source_xml.xpath("/pretext/*[not(docinfo)]//interactive")) > 0:
        image_output = (output / "preview").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating preview images for interactives\n\n")
        with utils.working_directory(Path()):
            try:
                core.preview_images(
                    xml_source=ptxfile,
                    pub_file=pub_file.as_posix(),
                    stringparams=params,
                    xmlid_root=xmlid_root,
                    dest_dir=image_output.as_posix(),
                )
            except Exception as e:
                log.error(
                    "Failed to generate interactive element previews. Check debug log for info."
                )
                log.debug(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.warning("Continuing...")
    else:
        log.info("Note: No interactive elements found.")


# generate youtube thumbnail assets


def youtube(ptxfile: Path, pub_file: Path, output: Path, params, xmlid_root):
    # We assume passed paths are absolute.
    # parse source so we can check for videos.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if len(source_xml.xpath("/pretext/*[not(docinfo)]//video[@youtube]")) > 0:
        image_output = (output / "youtube").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating youtube previews\n\n")
        with utils.working_directory(Path()):
            try:
                core.youtube_thumbnail(
                    xml_source=ptxfile,
                    pub_file=pub_file.as_posix(),
                    stringparams=params,
                    xmlid_root=xmlid_root,
                    dest_dir=image_output.as_posix(),
                )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error(
                    "Failed to generate some youtube video previews. Check your source and partial output to diagnose the issue."
                )
                log.warning("Continuing...")
    else:
        log.info("Note: No video@youtube elements found.")


# generate webwork assets


def webwork(ptxfile: Path, pub_file: Path, output: Path, params, xmlid_root=None):
    # We assume passed paths are absolute.
    # parse source so we can check for webwork.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if len(source_xml.xpath("//webwork[node()|@*]")) > 0:
        ww_output = (output / "webwork").resolve()
        os.makedirs(ww_output, exist_ok=True)
        log.info("Now generating webwork representation\n\n")
        with utils.working_directory(Path()):
            try:
                core.webwork_to_xml(
                    xml_source=ptxfile,
                    pub_file=pub_file.as_posix(),
                    stringparams=params,
                    xmlid_root=xmlid_root,
                    abort_early=True,
                    dest_dir=ww_output.as_posix(),
                    server_params=None,
                )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error(
                    "Failed to generate the webwork-representations file.  Check your source and partial output to diagnose the issue."
                )
                log.warning("Continuing...")
    else:
        log.info("Note: No webwork elements found.")


# generate codelens trace assets


def codelens(ptxfile: Path, pub_file: Path, output: Path, params, xmlid_root):
    # We assume passed paths are absolute.
    # parse source so we can check for webwork.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if len(source_xml.xpath("//program[@interactive = 'codelens']")) > 0:
        trace_output = (output / "trace").resolve()
        os.makedirs(trace_output, exist_ok=True)
        log.info("Now generating codelens trace\n\n")
        with utils.working_directory(Path()):
            try:
                core.tracer(
                    xml_source=ptxfile,
                    pub_file=pub_file.as_posix(),
                    stringparams=params,
                    xmlid_root=xmlid_root,
                    dest_dir=trace_output.as_posix(),
                )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error("Failed to generate codelens trace.")
                log.warning("Continuing...")
    else:
        log.info("Note: No program elements using codelens found.")


# generate youtube thumbnail assets


def qrcodes(ptxfile: Path, pub_file: Path, output: Path, params, xmlid_root):
    # We assume passed paths are absolute.
    # parse source so we can check for videos.
    source_xml = ET.parse(ptxfile)
    for _ in range(20):
        source_xml.xinclude()
    if (
        len(source_xml.xpath("/pretext/*[not(docinfo)]//video[@youtube]")) > 0
        or len(source_xml.xpath("/pretext/*[not(docinfo)]//interactive")) > 0
    ):
        image_output = (output / "qrcode").resolve()
        os.makedirs(image_output, exist_ok=True)
        log.info("Now generating QR codes for videos and/or interactives\n\n")
        with utils.working_directory(Path()):
            try:
                core.qrcode(
                    xml_source=ptxfile,
                    pub_file=pub_file.as_posix(),
                    stringparams=params,
                    xmlid_root=xmlid_root,
                    dest_dir=image_output.as_posix(),
                )
            except Exception as e:
                log.error(e)
                log.debug("Exception info:\n##################\n", exc_info=True)
                log.info("##################")
                log.error(
                    "Failed to generate some QR codes for some videos or interactives. Check your source and partial output to diagnose the issue."
                )
                log.warning("Continuing...")
    else:
        log.info("Note: No elements needing QR codes found")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pretext',
 'pretext.config',
 'pretext.core',
 'pretext.templates',
 'pretext.templates.resources']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3,<4',
 'PyPDF2>=2.5,<2.6',
 'click-log>=0.4,<0.5',
 'click>=8,<9',
 'ghp-import>=2,<3',
 'lxml>=4.8,<5.0',
 'pdfCropMargins>=1.0.9,<1.1.0',
 'playwright>=1,<2',
 'qrcode>=7,<8',
 'requests>=2,<3',
 'single-version>=1,<2',
 'watchdog>=2,<3']

entry_points = \
{'console_scripts': ['pretext = pretext.cli:main']}

setup_kwargs = {
    'name': 'pretext',
    'version': '1.4.1.dev20230202',
    'description': 'A package to author, build, and deploy PreTeXt projects.',
    'long_description': '# PreTeXt-CLI\n\nA package for authoring and building [PreTeXt](https://pretextbook.org) documents.\n\n-   GitHub: <https://github.com/PreTeXtBook/pretext-cli/>\n\n## Documentation and examples for authors/publishers\n\nMost documentation for PreTeXt authors and publishers is available at:\n\n-   <https://pretextbook.org/doc/guide/html/>\n\nAuthors and publishers may also find the examples catalog useful as well:\n\n-   <https://pretextbook.org/examples.html>\n\nWe have a few notes below (TODO: publish these in the Guide).\n\n### Installation\n\n#### Installing Python\n\nPreTeXt-CLI requires the Python version specified in `pyproject.toml`.\n\nTo check your version, type this into your terminal or command prompt:\n\n```\npython -V\n```\n\nIf your version is 2.x, try this instead\n(and if so, replace all future references to `python`\nin these instructions with `python3`).\n\n```\npython3 -V\n```\n\nIf you don\'t have a compatible Python available, try one of these:\n\n-   https://www.python.org/downloads/\n    -   Windows warning: Be sure to select the option adding Python to your Path.\n-   https://github.com/pyenv/pyenv#installation (Mac/Linux)\n-   https://github.com/pyenv-win/pyenv-win#installation (Windows)\n\n#### Installing PreTeXt-CLI\n\nOnce you\'ve confirmed that you\'re using a valid version of Python, just\nrun (replacing `python` with `python3` if necessary):\n\n```\npython -m pip install --user pretext\n```\n\n(It\'s possible you will get an error like\n`error: invalid command \'bdist_wheel\'`\n— good news, you can ignore it!)\n\nAfter installation, try to run:\n\n```\npretext --help\n```\n\nIf that works, great! Otherwise, it likely means that Python packages\naren\'t available on your “PATH”. In that case, replace all `pretext`\ncommands with `python -m pretext` instead:\n\n```\npython -m pretext --help\n```\n\nEither way, you\'re now ready to use the CLI, the `--help` option will explain how to use all the different\nsubcommands like `pretext new` and `pretext build`.\n\n#### External dependencies\n\nWe install as much as we can with the `pip install` command, but depending on your machine\nyou may require some extra software:\n\n-   [TeXLive](https://www.tug.org/texlive/)\n-   [pdftoppm/Ghostscript](https://github.com/abarker/pdfCropMargins/blob/master/doc/installing_pdftoppm_and_ghostscript.rst)\n\n#### Upgrading PreTeXt-CLI\n\nIf you have an existing installation and you want to upgrade to a more recent version, you can run:\n\n```\npython -m pip install --upgrade pretext\n```\n\n#### Custom XSL\n\nCustom XSL is not encouraged for most authors, but (for example) developers working\nbleeding-edge XSL from core PreTeXt may want to call XSL different from that\nwhich is shipped with a fixed version of the CLI. This may be accomplished by\nadding an `<xsl/>` element to your target with a relative (to `project.ptx`) or\nabsolute path to the desired XSL. _(Note: this XSL must only import\nother XSL files in the same directory or within subdirectories.)_\n\nFor example:\n\n```\n<target name="html">\n  <format>html</format>\n  <source>source/main.ptx</source>\n  <publication>publication/publication.ptx</publication>\n  <output-dir>output/html</output-dir>\n  <xsl>../pretext/xsl/pretext-html.xsl</xsl>\n</target>\n```\n\nIf your custom XSL file needs to import the XSL\nshipped with the CLI (e.g. `pretext-common.xsl`), then use a `./core/`\nprefix in your custom XSL\'s `xsl:import@href` as follows:\n\n```\n<xsl:import href="./core/pretext-common.xsl"/>\n```\n\nSimilarly, `entities.ent` may be used:\n\n```\n<!DOCTYPE xsl:stylesheet [\n    <!ENTITY % entities SYSTEM "./core/entities.ent">\n    %entities;\n]>\n```\n\n_Note: previously this was achieved with a `pretext-href` attribute - this is now deprecated and will be removed in a future release._\n\n---\n\n## Development\n\n**Note.** The remainder of this documentation is intended only for those interested\nin contributing to the development of this project. Anyone who simply wishes to\n_use_ the PreTeXt-CLI can stop reading here.\n\nFrom the "Clone or Download" button on GitHub, copy the `REPO_URL` into the below\ncommand to clone the project.\n\n```bash\ngit clone [REPO_URL]\ncd pretext-cli\n```\n\n### Using a valid Python installation\n\nDevelopers and contributors must install a\nversion of Python that matching the requirements in `pyproject.toml`.\n\n#### Using pyenv and poetry (Mac/Linux)\n\nThe `pyenv` tool for Linux automates the process of running the correct\nversion of Python when working on this project (even if you have\nother versions of Python installed on your system).\n\n-   https://github.com/pyenv/pyenv#installation\n\nRun the following, replacing `PYTHON_VERSION` with your desired version.\n\n```\npyenv install PYTHON_VERSION\n```\n\nThen follow these instructions to install `poetry`.\n\n-   https://python-poetry.org/docs/#installation\n    -   Note 2022/06/21: you may ignore "This installer is deprecated". See\n        [python-poetry/poetry/issues/4128](https://github.com/python-poetry/poetry/issues/4128)\n\nThen you should be able to install dependencies into a virtual environment\nwith this command.\n\n```\npoetry install\n```\n\nBefore you attempt to run `pretext` locally, you must fetch a copy of the core pretext library\nand zip up templates by running\n\n```\npoetry run python scripts/fetch_core.py\npoetry run python scripts/zip_templates.py\n```\n\nThen to use the in-development package, you can either enter a poetry shell:\n\n```\npretext --version # returns system version\npoetry shell\npretext --version # returns version being developed\nexit\npretext --version # returns system version\n```\n\nOr use the runner (as long as you remain within the package directory):\n\n```\npretext --version             # returns system version\npoetry run pretext --version  # returns version being developed\n```\n\nIf you run `echo \'alias pr="poetry run"\' >> ~/.bashrc` then restart your\nshell, this becomes less of a mouthful:\n\n```\npretext --version     # returns system version\npr pretext --version  # returns version being developed\n```\n\n#### Steps on Windows\n\nIn windows, you can either use the bash shell and follow the directions above,\nor try [pyenv-win](https://github.com/pyenv-win/pyenv-win#installation). In\nthe latter case, make sure to follow all the installation instructions, including\nthe **Finish the installation**. Then proceed to follow the directions above to\ninstall a version of python matching `pyproject.toml`. Finally, you may then need\nto manually add that version of python to your path.\n\n### Updating dependencies\n\nTo add dependencies for the package, run\n\n```\npoetry add DEPENDENCY-NAME\n```\n\nIf someone else has added a dependency:\n\n```\npoetry install\n```\n\n### Syncing untracked updates\n\nUpdates to certain files tracked to the repository will\nneed to be rebuilt by each user when pulled from GitHub.\n\nThe file `pretext/__init__.py` tracks the upstream\ncommit of core PreTeXt XSL/Python code we\'re developing against\n(from `PreTeXtBook/pretext`).\nTo fetch these updates from upstream, run:\n\n```\npoetry run python scripts/fetch_core.py\n```\n\nIf you instead want to point to a local copy of `PreTeXtBook/pretext`,\ntry this instead to set up symlinks:\n\n```\npoetry run python scripts/symlink_core.py path/to/pretext\n```\n\nUpdates to `templates/` must be zipped and moved into\n`pretext/templates/resources`. This is done automatically by\nrunning:\n\n```\npoetry run python scripts/zip_templates.py\n```\n\n### Formatting code before a commit\n\nAll `.py` files are formatted with the [black](https://black.readthedocs.io/en/stable/)\npython formatter and checked by [flake8](https://flake8.pycqa.org/en/latest/).\nProper formatting is enforced by checks in the Continuous Integration framework.\nBefore you commit code, you should make sure it is formatted with `black` and\npasses `flake8` by running the following commands (on linux or mac)\nfrom the _root_ project folder (most likely `pretext-cli`).\n\n```\npoetry run black .\npoetry run flake8\n```\n\n### Testing\n\nSets are contained in `tests/`. To run all tests:\n\n```\npoetry run pytest\n```\n\nTo run a specific test, say `test_name` inside `test_file.py`:\n\n```\npoetry run pytest -k name\n```\n\nTests are automatically run by GitHub Actions when pushing to identify\nregressions.\n\n### Packaging\n\nTo check if a successful build is possible:\n\n```\npoetry run python scripts/build_package.py\n```\n\nTo publish a new alpha release, first add/commit any changes. Then\nthe following handles bumping versions, publishing to PyPI,\nand associated Git management.\n\n```\npoetry run python scripts/release_alpha.py\n```\n\nPublishing a stable release is similar:\n\n```\npoetry run python scripts/release_stable.py # patch +0.+0.+1\npoetry run python scripts/release_stable.py minor # +0.+1.0\npoetry run python scripts/release_stable.py major # +1.0.0\n```\n\n---\n\n## About\n\n### PreTeXt-CLI Team\n\n-   [Oscar Levin](https://math.oscarlevin.com/) is co-creator and lead developer of PreTeXt-CLI.\n-   [Steven Clontz](https://clontz.org/) is co-creator and a regular contributor of PreTeXt-CLI.\n-   Development of PreTeXt-CLI would not be possible without the frequent\n    [contributions](https://github.com/PreTeXtBook/pretext-cli/graphs/contributors) of the\n    wider [PreTeXt-Runestone Open Source Ecosystem](https://prose.runestone.academy).\n\n### A note and special thanks\n\nA `pretext` package unrelated to the PreTeXtBook.org project was released on PyPI\nseveral years ago by Alex Willmer. We are grateful for his willingness to transfer\nthis namespace to us.\n\nAs such, versions of this project before 1.0 are released on PyPI under the\nname `pretextbook`, while versions 1.0 and later are released as `pretext`.\n\n### About PreTeXt\n\nThe development of [PreTeXt\'s core](https://github.com/PreTeXtBook/pretext)\nis led by [Rob Beezer](http://buzzard.ups.edu/).\n',
    'author': 'Oscar Levin',
    'author_email': 'oscar.levin@unco.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pretextbook.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.5,<4.0.0',
}


setup(**setup_kwargs)



from pathlib import Path
import zipfile
import importlib.resources
from .. import CORE_COMMIT


def path(*args) -> Path:
    # Checks that the local static path ~/.ptx/ contains the static files needed for core, and installs them if they are missing (or if the version is different from the installed version of pretext).  Then returns the absolute path to the static files (appending arguments)
    local_base_path = Path.home() / ".ptx"
    local_commit_file = Path(local_base_path) / ".commit"
    if not Path.is_file(local_commit_file):
        print("Static pretext files do not appear to be installed.  Installing now.")
        install(local_base_path)
    # check that the static core_commit matches current core_commit
    with open(local_commit_file, "r") as f:
        static_commit = f.readline().strip()
    if static_commit != CORE_COMMIT:
        print("Static pretext files are out of date.  Installing them now.")
        install(local_base_path)
    return local_base_path.joinpath(*args)


def install(local_base_path):
    with importlib.resources.path("pretext.core", "resources.zip") as static_zip:
        with zipfile.ZipFile(static_zip, "r") as zip:
            zip.extractall(local_base_path)
    # Write the current commit to local file
    with open(local_base_path / ".commit", "w") as f:
        f.write(CORE_COMMIT)
    print(
        f"Static files required for pretext have now been installed to {local_base_path}"
    )
    return



import click
import typing as t
from lxml import etree as ET
from ..utils import format_docstring_as_help_str


USAGE_DESCRIPTION = format_docstring_as_help_str(
    """
    Override an entry in the project.ptx file. Entries are specified as `path` and `value`.
    Paths are `.`-separated. For example, the path of the `<c>` node in `<a><b><c>...` is
    `a.b.c`.

    Attributes can be overridden providing `path@attribute_name` and `value`. For example,
    to override the `foo` attribute in `<a><b><c foo="bar">...` use `a.b.c@foo`.

    Example, to over add/override custom xsl, you could run
    `pretext build {} targets.target.xsl custom.xsl`
"""
)


class ShadowXmlNodeType(t.TypedDict):
    name: t.Union[str, None]
    value: t.Union[str, None]
    attributes: t.List[t.Tuple[str, str]]


class XmlOverlayType(click.ParamType):
    name = "xml_overlay"

    def convert(self, value, param, ctx):
        print("got", value, param, ctx.args, ctx.obj)
        return value


class ShadowXmlDocument:
    def __init__(self):
        self._nodes_dict: t.Dict[str, ShadowXmlNodeType] = {}

    def upsert_node_or_attribute(self, path: str, value: t.Union[str, None]):
        """
        Upserts a node into the shadow document.

        @path - A string representing the path of the node/attribute. For example, `a.b.c` would be nested `<a><b><c>...` nodes.
                `a.b@c` would be an attribute named `c` on the `b` node. E.g. `<a><b c="...">...`
        @value - The (string) value of the node/attribute.
        """

        # paths of the form a.b.c@foo represent the attribute `foo` on the `c` node
        if "@" in path:
            if len(path.split("@")) > 2:
                raise Exception(
                    "Cannot have multiple `@` characters in a ShadowXmlDocument path."
                )
            path, attr_name = path.split("@")

            node = self._nodes_dict.get(
                path, {"name": None, "value": None, "attributes": []}
            )
            node["attributes"].append((attr_name, value))
            self._nodes_dict[path] = node
            # all other paths correspond to nodes and values
        else:
            node = self._nodes_dict.get(
                path, {"name": None, "value": None, "attributes": []}
            )
            node["value"] = value
            self._nodes_dict[path] = node
        return self

    def overlay_tree(self, root: ET.Element = ET.Element("root")) -> t.List[str]:
        """
        Overlay `root` with the current ShadowXmlDocument's nodes and attributes.
        A list of string messages are returned about what elements were changed.

        This function mutates `root`.
        """

        ret: t.List[str] = []

        def upsert_node(
            path: t.List[str],
            current: ET.Element = root,
            current_path: t.List[str] = [],
        ) -> t.List[ET.Element]:
            if len(path) == 0:
                return [current]
            needed_tag = path[0]
            path = path[1:]
            current_path = [*current_path, needed_tag]

            # Find all matching tags. If none are found, we create one.
            # If any are found we recurse down _every_ instance of a found tag
            matching_elms = [elm for elm in current if elm.tag == needed_tag]
            if len(matching_elms) == 0:
                # if we're here, there was no node with the correct tag name, so we add one.
                new_node = ET.Element(needed_tag)
                current.append(new_node)
                ret.append("NODE_ADDED with XML path {}".format(".".join(current_path)))
                return upsert_node(path, new_node, current_path)

            # There was one or more tags with the correct name;
            # Concatenate all target nodes that are found.
            return sum(
                (upsert_node(path, node, current_path) for node in matching_elms), []
            )

        for path, node in sorted(self._nodes_dict.items()):
            split_path = path.split(".")
            xml_nodes = upsert_node(split_path)
            for xml_node in xml_nodes:
                for attr, val in node["attributes"]:
                    old_val = xml_node.get(attr)
                    if old_val is not None:
                        if old_val != val:
                            ret.append(
                                f"ATTRIBUTE_CHANGED '{old_val}' to '{val}' at XML path '{path}'"
                            )
                    else:
                        ret.append(f"ATTRIBUTE_ADDED '{val}' at XML path '{path}'")
                    xml_node.set(attr, val)
                if node["value"] is not None:
                    if xml_node.text is not None:
                        ret.append(
                            "TEXT_CHANGED '{}' to '{}' at XML path '{}'".format(
                                repr(xml_node.text), repr(node["value"]), path
                            )
                        )
                    else:
                        ret.append(
                            "TEXT_ADDED '{}' at XML path '{}'".format(
                                repr(node["value"]), path
                            )
                        )
                    xml_node.text = node["value"]

        return ret

    def __repr__(self) -> str:
        node_repr = []
        for path, node in self._nodes_dict.items():
            attrs = " ".join(f'{key}="{val}"' for key, val in node["attributes"])
            value = "" if node["value"] is None else node["value"]
            node_repr.append(f"""<{path} {attrs}>{value}</>""")
        return "ShadowXmlDocument:" + "\n".join(sorted(node_repr))
